/*
  Tool            zz_log
  Author          Carlo Hogeveen
  Website         eCarlo.nl/tse
  Compatibility   TSE 2.5e upwards, all TSE variants
  Version         1   1 Oct 2022

  zz_log implements old-school debugging!

  If a target macro source file is current, then this tool adds zz_log lines,
  that, when the target macro is run, log what parts of the target macro are
  actually executing.

  You can add your own zz_log_write() lines.

  Afterwards all zz_log lines are easily removed with TSE's Find&Do menu.


  INSTALLATION

  Copy zz_log.s and zz_log.inc to TSE's "mac" directory and compile "zz_log.s"
  there, for example by opening it in TSE and applying the Macro Compile menu.


  USAGE

  Open the the source file of the target macro you want to debug.
  Select TSE's Macro Execute menu and execute "zz_log".

  Compile the target macro and resolve any compilation errors:
    zz_log is good at placing its zz_log statements, but not perfect.
    The compilation will reveal this as usually obvious and easily-remedied
    errors.

  Check whether the zz_log_close() state is placed (correctly).
  zz_log cannot always determine when not to place it at the end of the Main()
  proc, or where to place it if there are no Main() and WhenPurged() proc.

  After running the target macro, you might need to Macro Purge it,
  in order to make the contents of the log file appear non-empty.

  In Linux the log file will be written to the "/tmp" directory.
  In Windows the log file will be written to the directory from the user's TMP
  environment variable, usually C:\Users\<UserName>\appdata\local\temp.

  The name of the log file will be:
    tse_zz_log_<name of target macro>_<day's time in 1/100 seconds>.log

  You can add your own zz_log_write() statements.
  With zz_log_write(format( ... )) you can make them fancy and detailed.
  Do make them one-liners for easy removal later on.

  After you are done debugging, you can simply delete all zz_log lines
  from the target macro, by using TSE's Search Find&Do menu with
    Search for: zz_log
    Options: gin
    Delete Line

  Happy debugging!

*/





// Start of compatibility restrictions and mitigations


// The following determines the editor version at compile-time.
#ifdef EDITOR_VERSION
  // From TSE 3.0 upwards the built-in constant EDITOR_VERSION exists.
#else
  #ifdef WIN32
    // From TSE 2.6 upwards the built-in constant WIN32 exists.
    #define EDITOR_VERSION 0x2800
    // TSE versions < 2.5e and TSE 2.6 are not supported by this macro.
    // TSE 2.5e (the Dos version of TSE) has existed so long, that almost
    // every Dos user has upgraded to it, so supporting older versions is
    // no longer worth the effort. Due to a bug in TSE 2.5c (and presumably
    // lower versions, which is one of the reasons for not supporting them)
    // the line below is syntax-checked despite the lack of WIN32 in those
    // versions, and generates a desired compiler error.
    // TSE 2.6 was an intermediate 32-bits version from which anyone could
    // freely upgrade to TSE 2.8. The parameter constant _MFSKIP_ exists
    // since TSE 2.8, so the line below will also generate a desired
    // compiler error for TSE 2.6.

    // TSE versions < 2.5e and TSE 2.6 are not supported by this macro.
    #define UNSUPPORTED_TSE_VERSION _MFSKIP_
  #else
    // Since all other TSE versions were excluded, this must be TSE 2.5e.
    #define EDITOR_VERSION 0x2500
  #endif
#endif


#if EDITOR_VERSION == 0x2500
  #define MAXSTRINGLEN 255
#endif


// End of compatibility restrictions and mitigations.





proc process_a_dll()
  // Just skip till the end.
  lFind('^ *end |$', 'ix+')
end process_a_dll


proc process_a_multiline_comment()
  // Just skip till the end.
  lFind('*/', '+')
end process_a_multiline_comment


integer proc find_first_non_data_definition()
  integer found       = FALSE
  integer indentation = 0

  Down()
  repeat
    indentation = iif(PosFirstNonWhite() > 1, PosFirstNonWhite() - 1,
                                              indentation)
    if not (  lFind('^ *$', 'cgx')
           or lFind('^ */|\#', 'cgx')
           or lFind('^ *{integer}|{string} ', 'cgix'))
      found = TRUE
    elseif lFind('/\*', 'cgx')
      lFind('\*/', 'x+')
    endif
  until found
     or not Down()

  if not found
    AddLine() // Add a line above which to insert.
  endif

  return(indentation)
end find_first_non_data_definition


proc process_a_proc()
  integer end_found                 = FALSE
  integer indentation               = 0
  string  proc_name  [MAXSTRINGLEN] = ''
  integer is_proc_with_return_value = FALSE

  if (Lower(GetWord()) in 'integer', 'string')
    is_proc_with_return_value = TRUE
    WordRight()
  endif

  WordRight()
  proc_name = GetWord()

  // Insert "Begin" logging.
  indentation = find_first_non_data_definition()
  InsertLine(Format('': indentation, "zz_log_write('B ", proc_name, "')"))

  // Insert "Return" and "End" logging.
  Down()
  repeat
    indentation = iif(PosFirstNonWhite() > 1, PosFirstNonWhite() - 1,
                                              indentation)
    if lFind('^ *end |$', 'cgix')
      if not is_proc_with_return_value
        InsertLine(Format('': indentation, "zz_log_write('E ", proc_name, "')"))
      endif
      end_found = TRUE
    elseif lFind('^ *return |(|$', 'cgix')
      InsertLine(Format('': indentation, "zz_log_write('R ", proc_name, "')"))
      Down()
    elseif lFind('/\*', 'cgx')
      lFind('\*/', 'x+')
    endif
  until end_found
     or not Down()
end process_a_proc


proc add_zz_log_open()
  integer indentation = 0

  if lFind('^ *proc #WhenLoaded *(', 'gix')
  or lFind('^ *proc #Main *('      , 'gix')
    indentation = find_first_non_data_definition()
    InsertLine(Format('': indentation, 'zz_log_open()'))
  endif
end add_zz_log_open


proc add_zz_log_close()
  integer found       = FALSE
  integer indentation = 0

  if lFind('^ *proc #WhenPurged *(', 'gix')
  or lFind('^ *proc #Main *('      , 'gix')

    // Find the proc's "end" and indentation, skipping multi-line comments.
    Down()
    repeat
      indentation = iif(PosFirstNonWhite() > 1, PosFirstNonWhite() - 1,
                                                indentation)
      if lFind('^ *end |$', 'cgix')
        found = TRUE
      elseif lFind('/\*', 'cgx')
        lFind('\*/', 'x+')
      endif
    until found
       or not Down()

    if not found
      AddLine()
    endif

    InsertLine(Format('': indentation, 'zz_log_close()'))
  endif
end add_zz_log_close


proc add_zz_log_lines()
  string search_options [3] = 'gix'

  BegFile()
  InsertLine("#Include ['zz_log.inc']")

  while lFind('^ *\c{dll}|{{integer #}|{string #}|{}proc }|{/\*}',
              search_options)
    if Lower(GetText(1, 3)) == 'dll'
      process_a_dll()
    elseif GetText(1, 2) == '/*'
      process_a_multiline_comment()
    else
      process_a_proc()
    endif
    search_options = 'ix+'
  endwhile

  add_zz_log_close()
  add_zz_log_open()

  BegFile()
  UpdateDisplay(_ALL_WINDOWS_REFRESH_)
  lFind('zz_log', 'giv')
end add_zz_log_lines


proc Main()
  if (CurrExt() in '.s', '.si')
    add_zz_log_lines()
  else
    Warn('The current file is not a macro source!')
  endif
  PurgeMacro(CurrMacroFilename())
end Main

